// Copyright 1995 by Jon Dart.  All Rights Reserved.

// Utility to add ECO codes to a PGN data file

#include "stdafx.h"
#include "board.h"
#include "movegen.h"
#include "options.h"
#include "movearr.h"
#include "notation.h"
#include "chessio.h"
#include "eco.h"
#include "movearr.h"
#include "rmove.h"
extern "C"
{
#include <string.h>
#include <ctype.h>
};
#include <fstream.h>
#include <iostream.h>
#include <strstrea.h>
#include <ctype.h>

Options *global_options = NULL;
Move_Array *game_moves = NULL;
BOOL replace = TRUE;

const MARGIN = 75;

static void show_usage()
{
	 cerr << "Usage: ecocoder pgn_file" << endl;
}

static int legal_move(const Board &board, const Square &start,
  const Square &dest)
{
    // do a little basic sanity checking:
      if (!start.OnBoard() || !dest.OnBoard() || board[start].IsEmpty())
          return False;

      ExtendedMove emove(board,start,dest);
      Move moves[Move_Generator::MaxMoves];
      Move_Generator mg( board, 0, Move::NullMove() );
      int found = 0;
          int n = mg.Generate_Moves(moves, False);

      for (int i = 0; i < n; i++)
          {
          if (moves[i].StartSquare() == emove.StartSquare() &&
              moves[i].DestSquare() == emove.DestSquare())
          {
             found++;
             break;
          }
      }
      if (!found)
          return False;
          else        
      {
          // check for king en prise
          Board board_copy(board);
          const ColorType side = board.Side();
          board_copy.MakeMove(emove);
          return board_copy.num_attacks(
            board_copy.KingPos(side),OppositeColor(side)) == 0;
          
      }
}

void write_pgn(ostream &game_file, ChessIO::hdr *hdrs, 
                const Move_Array &moves, const CString &eco,
                const CString result)
{
    int eco_found = 0;
    ChessIO::hdr *h = hdrs;
    while (h)
    {
        int skip = 0;
	if (strcmp(h->tag,"ECO")==0)
	{
	   eco_found++;
// start debug
	   if (strcmp(h->val,(LPCSTR)eco)!=0)
	     cerr << "warning, ECO codes differ: " << h->val <<
	     	" " << (LPCSTR)eco << endl;
// end debug	   
	   if (replace)
	   {
              game_file << "[ECO \"" << (LPCSTR)eco << "\"]" << endl;
	      ++skip;
	   }
	}
	if (!skip)
            game_file << '[' << h->tag << ' ' << '"' << h->val << '"' << ']' << endl;
	h = h->next;
    }
    if (*eco && !eco_found)
       game_file << "[ECO \"" << eco << "\"]" << endl;
    hdrs = NULL;
    game_file << endl;
    Board board;
    int max = moves.num_moves();
    char buf[80];
    *buf = '\0';
    char num[10];
    int last = 0;
    int num_image_size;
    for (int i = 0; i < max; i++)
    {
       Move_Record mr = moves[i];
       num_image_size = 0;
       *num = '\0';
       if (i % 2 == 0)
       {
	   sprintf(num,"%d. ",(i/2)+1);
	   num_image_size = strlen(num);
       }
       char img[20];
       Notation::Image(board,mr.move(),img);
       const int image_size = strlen(img);
       if (last + image_size + num_image_size + 1 >= MARGIN)
       {
	    buf[last] = '\0';
	    game_file << buf << endl;
    	    last = 0;
   	    *buf = '\0';
       }
       if (last != 0)
	    strcat(buf," ");
       strcat(buf,num);
       strcat(buf,img);
       last = strlen(buf);
       board.MakeMove(mr.move());
    }
    if (max)
    {
    if (last)
    {	
           if (last + strlen(result) + 1 >= MARGIN)
	   {
              buf[last] = '\0';
	      game_file << buf << endl;
      	      game_file << (LPCSTR)result << endl;
	   }
	   else
	   {
	       strcat(buf," ");
	       strcat(buf,(LPCSTR)result);
	       buf[last + strlen((LPCSTR)result) + 1] = '\0';
	       game_file << buf << endl;
	   }
    }
    else
    	game_file << (LPCSTR)result << endl;
    }
    game_file << endl;
}

int main(int argc, char **argv)
{
   Board board;
   ECO eco;
   int arg = 1;

   if (argc ==1)
   {
         show_usage();
	 return -1;
   }
   else
   {
	ifstream pgn_file( argv[arg], ios::in);
        CString result;
	int c;
        ColorType side;
	if (pgn_file.good())
	{
            ChessIO::hdr *hdrs = NULL;
           
            while (!pgn_file.eof())
            {
               long first;
               Move_Array moves;
               board.Reset();
               side = White;
               while (hdrs)
               {
                  ChessIO::hdr *next = hdrs->next;
                  delete hdrs;
                  hdrs = next;
               }
               hdrs = NULL;
               while (pgn_file.good() && (c = pgn_file.get()) != EOF)
               {
                   if (c=='[')
                   {
                      pgn_file.putback(c);
                      break;
                   }
               }
               ChessIO::collect_headers(pgn_file,hdrs,first);
               for (;;)
               {
                   static char buf[2048];
                   static char num[20];
                   ChessIO::Token tok = ChessIO::get_next_token(pgn_file,buf,2048);
                   if (tok == ChessIO::Eof)
                      break;
                   if (tok == ChessIO::Number)
                   {
                      strncpy(num,buf,20);
                   }
                   else if (tok == ChessIO::GameMove)
                   {
                      // parse the move
                      Move m = Notation::Value(board,side,buf);
                      if (m.IsNull() ||
                        !legal_move(board,m.StartSquare(),
                            m.DestSquare()))
                      {
                          // echo to both stdout and stderr
                          fprintf(stderr,"Illegal move: %s\n",buf);
                          cout << "Illegal move: " << buf << endl;
                          break;
                      }
                      else
                      {
                          /**
                          char result[40];
                          Notation::Image(board,m,result); **/
                          ExtendedMove emove(board,m);
                          moves.add_move(board,emove);
                          
                          ReversibleMove rmove(board,emove);
                          board.MakeMove(rmove);
                          // cout << result << endl;
                          
                      }
                      side = OppositeColor(side);
                  }
                  else if (tok == ChessIO::Unknown)
                  {
                      fprintf(stderr,"Unrecognized text: %s\n",buf);
                  }
                  else if (tok == ChessIO::Comment)
                  {
                      // ignored for now
                  }
                  else if (tok == ChessIO::Result)
                  {
                       result = buf;
                       break;
                  }
               }
               // done with a game
               CString ecoCode, name;
               eco.classify(moves,ecoCode,name);
               write_pgn(cout,hdrs,moves,ecoCode,result);
            }
            pgn_file.close();
	}
	else
	{
  	    cerr << "file not found: " << argv[arg] << endl;
	    return -1;
	}
   }

   return 0;
}
